"""project_id

Revision ID: be206abd1412
Revises: f894faeb055d
Create Date: 2018-03-29 22:42:42.636390

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'be206abd1412'
down_revision = 'f894faeb055d'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tProject', 'project_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tProject', 'project_name', new_column_name='name', existing_type=sa.String(200))
            op.alter_column('tProject', 'project_description', new_column_name='description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        # ## tProject
        try:
            op.drop_table('tProject_new')
        except:
            log.info("tProject_new isn't there")

        try:
            # ## tProject
            op.create_table(
                'tProject_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id'),  nullable=False),
                sa.UniqueConstraint('name', 'created_by', 'status', name="unique project name")
            )

            op.execute("insert into tProject_new (id, name, description, status, cr_date, created_by) select project_id, project_name, project_description, status, cr_date, created_by from tProject")

            op.rename_table('tProject','tProject_old')
            op.rename_table('tProject_new', 'tProject')
            op.drop_table('tProject_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tProject

        try:
            op.alter_column('tProject', 'id', new_column_name='project_id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tProject', 'name', new_column_name='project_name', existing_type=sa.String(200))
            op.alter_column('tProject', 'description', new_column_name='project_description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)

    else: ## sqlite
        try:
            # ## tProject
            op.create_table(
                'tProject_new',
                sa.Column('id', sa.Integer, primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id'),  nullable=False),
                sa.UniqueConstraint('name', 'created_by', 'status', name="unique project name")
            )

            op.execute("insert into tProject_new (project_id, project_name, project_description, status, cr_date, created_by) select id, name, description, status, cr_date, created_by from tProject")

            op.rename_table('tProject','tProject_old')
            op.rename_table('tProject_new', 'tProject')
            op.drop_table('tProject_old')

        except Exception as e:
            log.exception(e)



"""template_status

Revision ID: 1d3ab26415ec
Revises: b04cb2e57cb0
Create Date: 2020-12-11 10:24:28.203040

"""

import logging
from alembic import op
import sqlalchemy as sa

log = logging.getLogger(__name__)


# revision identifiers, used by Alembic.
revision = '1d3ab26415ec'
down_revision = 'b04cb2e57cb0'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        try:
            op.add_column('tTemplate', sa.Column('status', sa.String(1), nullable=False, server_default=sa.text(u"'A'")))
        except Exception as e:
            log.critical(e)


def downgrade():
    if op.get_bind().dialect.name == 'mysql':
        try:
            op.drop_column('tTemplate', 'status')
        except Exception as e:
            log.critical(e)



"""
  Hydra Type Registry

  Collects Hydra Types (classes having DataType as their base) and
  constructs a `typemap` mapping the unique identifying str tag
  attribute of each type to the corresponding class.
"""
import json
import inspect
import sys

typemap = {}
from .Types import Array, Scalar, Timeseries, Descriptor, Dataframe
from .Types import DataType as Datatype_Base


class HydraObjectFactory(object):
    """
      Abstract class serving as a factory for the creation of Hydra Types
    """
    @staticmethod
    def fromJSON(encstr, tmap=None):
        if tmap is None:
            tmap = typemap
        pass

    @classmethod
    def valueFromDataset(cls, datatype, value, metadata=None, tmap=None):
        """
          Return the value contained by dataset argument, after casting to
          correct type and performing type-specific validation
        """
        if tmap is None:
            tmap = typemap
        obj = cls.fromDataset(datatype, value, metadata=metadata, tmap=tmap)
        return obj.value

    @staticmethod
    def fromDataset(datatype, value, metadata=None, tmap=None):
        """
          Return a representation of dataset argument as an instance

          of the class corresponding to its datatype
        """
        if tmap is None:
            tmap = typemap
        return tmap[datatype.upper()].fromDataset(value, metadata=metadata)




import pytest

import tempfile
import datetime
import sqlite3
import time

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

import hydra_base
from hydra_base.db import DeclarativeBase as _db
from hydra_base.util.hdb import create_default_users_and_perms, make_root_user,\
                                create_default_units_and_dimensions
from hydra_base.util import testing
from hydra_client.connection import JSONConnection, RemoteJSONConnection
from hydra_base.lib.cache import clear_cache


no_externaldb_opt = "--no-externaldb"
externaldb_mark = "externaldb"
requires_hdf_mark = "requires_hdf"

def pytest_addoption(parser):
    parser.addoption("--db-backend", action="store", default="sqlite",
                     help="Database backend to use when running the tests.")
    parser.addoption("--connection-type", action="store", default="local",
                     help="Remote or Local Connection")
    parser.addoption(no_externaldb_opt, action="store_true", default=False,
                     help="Do not run tests which require an external storage database")

def pytest_configure(config):
    config.addinivalue_line("markers", f"{externaldb_mark}: Tests external storage")
    config.addinivalue_line("markers", f"{requires_hdf_mark}: Indicates that decorated test requires HDF support")

def pytest_collection_modifyitems(config, items):
    """
    When the `no_externaldb_opt` is present, add a skip mark to every
    test marked with `externaldb_mark`

    When config has disabled HDF support, add a skip mark to every test
    marked with `requires_hdf_mark`
    """
    if config.getoption(no_externaldb_opt):
        externaldb_skip = pytest.mark.skip(reason=f"{no_externaldb_opt} selected")
        for item in items:
            if externaldb_mark in item.keywords:
                item.add_marker(externaldb_skip)

    conf_disabled = hydra_base.config.CONFIG.get("storage_hdf", "disable_hdf").lower()
    hdf_disabled = True if conf_disabled in ("true", "yes") else False
    hdf_skip = pytest.mark.skip(reason="Test not applicable when HDF support disabled")
    if hdf_disabled:
        for item in items:
            if requires_hdf_mark in item.keywords:
                item.add_marker(hdf_skip)


@pytest.fixture(scope="session")
def db_backend(request):
    return request.config.getoption("--db-backend")

@pytest.fixture(scope="session")
def connection_type(request):
    return request.config.getoption("--connection-type")

def pytest_report_header(config):
    headers = []
    db_string = config.getoption("--db-backend")
    headers.append('db-backend: {}'.format(db_string))
    connection_type = config.getoption("--connection-type")
    headers.append('connection-type: {}'.format(connection_type))
    return '\n'.join(headers)


"""
    Test Fixtures
"""
#def pytest_namespace():
#    return {'root_user_id': 1}

@pytest.fixture()
def dateformat():
    return hydra_base.config.get('DEFAULT', 'datetime_format', "%Y-%m-%dT%H:%M:%S.%f000Z")

@pytest.fixture(scope='module')
def testdb_uri(db_backend):
    if db_backend == 'sqlite':
        tmp = tempfile.gettempdir()
        # Use a :memory: database for the tests.
        millis = int(round(time.time() * 1000))
        return f'sqlite:///{tmp}/test_db_{millis}.db'
    elif db_backend == 'postgres':
        # This is designed to work on Travis CI
        return 'postgresql://postgres@localhost:5432/hydra_base_test'
    elif db_backend == 'mysql':
        return 'mysql+mysqldb://root:root@localhost/hydra_base_test'
    else:
        raise ValueError('Database backend "{}" not supported when running the tests.'.format(db_backend))

@pytest.fixture(scope='module')
def client(connection_type, testdb_uri):
    print("GETTING CLIENT")
    if connection_type == 'local':
        client = JSONConnection(app_name='Hydra Local Test Suite', db_url=testdb_uri)
        #fake a login using the test's session
        client.user_id = 1
        client.connect()
    else:

        from spyne.server.null import NullServer
        from hydra_server import initialize_api_server

        hydra_server = initialize_api_server(testdb_uri, test=True)

        null_server = NullServer(hydra_server.json_application, ostr=True)

        #The url argument here is to avoid the connection complaining.
        #It's not actually used, as we're using a null (testing) server
        client = RemoteJSONConnection(url='localhost:8080/json',
                                      app_name='Hydra Remote Test Suite',
                                      test_server=null_server)
        client.login('root', '')

    client.testutils = testing.TestUtil(client)
    pytest.root_user_id = 1
    pytest.user_a = client.testutils.create_user("UserA")
    pytest.user_b = client.testutils.create_user("UserB")
    pytest.user_c = client.testutils.create_user("UserC", role='developer')
    pytest.user_d = client.testutils.create_user("UserD", role='developer')
    yield client

    hydra_base.db.close_session()

    clear_cache() # clear the user project cache

    try:
        drop_tables(testdb_uri)
    except Exception as err:
        print("Error dropping DB")

def drop_tables(db_url):
    """
        Drop all the tables in the specified DB
    """
    from hydra_base.db import DeclarativeBase, engine
    DeclarativeBase.metadata.drop_all(engine)

@pytest.fixture()
def network(client, project_id=None, num_nodes=10, new_proj=True, map_projection='EPSG:4326'):
    return client.testutils.build_network(project_id, num_nodes, new_proj, map_projection)

@pytest.fixture()
def network_with_data(client, project_id=None, num_nodes=10, ret_full_net=True, new_proj=True, map_projection='EPSG:4326'):
    return client.testutils.create_network_with_data(project_id, num_nodes, ret_full_net, new_proj, map_projection)

@pytest.fixture()
def network_with_child_scenario(client, project_id=None, num_nodes=10, ret_full_net=True, new_proj=True, map_projection='EPSG:4326'):
    return client.testutils.create_network_with_child_scenario(project_id, num_nodes, ret_full_net, new_proj, map_projection, levels=2)

@pytest.fixture()
def network_with_grandchild_scenario(client, project_id=None, num_nodes=10, ret_full_net=True, new_proj=True, map_projection='EPSG:4326'):
    """
    Creates a network with 3 scenarios -- a baseline, a child, and a child of the child.
    """
    return client.testutils.create_network_with_child_scenario(project_id, num_nodes, ret_full_net, new_proj, map_projection, levels=3)

@pytest.fixture()
def second_network_with_data(client, project_id=None, num_nodes=10, ret_full_net=True, new_proj=True, map_projection='EPSG:4326'):
    """
        Creates a second project with a new network and returns the network created
    """
    return client.testutils.create_network_with_data(project_id, num_nodes, ret_full_net, new_proj, map_projection)

@pytest.fixture()
def network_with_extra_group(client, project_id=None, num_nodes=10, ret_full_net=True, new_proj=True, map_projection='EPSG:4326'):
    return client.testutils.create_network_with_extra_group(project_id, num_nodes, ret_full_net, new_proj, map_projection)


@pytest.fixture()
def networkmaker(client):
    class NetworkMaker:
        def create(self, project_id=None, num_nodes=10, ret_full_net=True, new_proj=True, map_projection='EPSG:4326'):
            return client.testutils.create_network_with_data(project_id, num_nodes, ret_full_net, new_proj, map_projection)
    return NetworkMaker()

@pytest.fixture()
def template(client):
    return client.testutils.create_template()

@pytest.fixture()
def attributes(client):
    return client.testutils.create_attributes()

@pytest.fixture()
def attribute(client):
    return client.testutils.create_attribute()


@pytest.fixture()
def projectmaker(client):
    class ProjectMaker:
        def create(self, name=None, share=True, parent_id=None):
            if name is None:
                name = 'Project %s' % (datetime.datetime.now())
            return client.testutils.create_project(name=name, share=share, parent_id=parent_id)

    return ProjectMaker()


@pytest.fixture()
def attributegroup(client):
    project = client.testutils.create_project('Project %s' % (datetime.datetime.now()))
    newgroup = client.testutils.create_attributegroup(project.id)
    return newgroup

@pytest.fixture()
def attributegroupmaker(client):
    class AttributeGroupMaker:
        def create(self, project_id, name=None, exclusive='N'):

            if name is None:
                name = 'Attribute Group %s' % (datetime.datetime.now())

            newgroup = client.testutils.create_attributegroup(project_id,
                                                  name=name,
                                                  exclusive=exclusive)

            return newgroup

    return AttributeGroupMaker()

@pytest.fixture()
def new_dataset(client):
    return client.testutils.create_dataset()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from ..db.model import Note
from .. import db
from ..exceptions import HydraError, ResourceNotFoundError
from sqlalchemy.orm.exc import NoResultFound

def _get_note(note_id):
    try:
        note_i = db.DBSession.query(Note).filter(Note.id == note_id).one()
    except NoResultFound:
        raise ResourceNotFoundError("Note %s not found"%note_id)
    return note_i

def get_notes(ref_key, ref_id, **kwargs):
    """
        Get all the notes for a resource, identifed by ref_key and ref_ido.
        Returns [] if no notes are found or if the resource doesn't exist.
    """
    notes = db.DBSession.query(Note).filter(Note.ref_key==ref_key)
    if ref_key == 'NETWORK':
        notes = notes.filter(Note.network_id == ref_id)
    elif ref_key == 'NODE':
        notes = notes.filter(Note.node_id == ref_id)
    elif ref_key == 'LINK':
        notes = notes.filter(Note.link_id == ref_id)
    elif ref_key == 'GROUP':
        notes = notes.filter(Note.group_id == ref_id)
    elif ref_key == 'PROJECT':
        notes = notes.filter(Note.project_id == ref_id)
    elif ref_key == 'SCENARIO':
        notes = notes.filter(Note.scenario_id == ref_id)
    else:
        raise HydraError("Ref Key %s not recognised.")

    note_rs = notes.all()

    return note_rs

def get_note(note_id, **kwargs):
    """
    Get a note by ID
    """
    note = _get_note(note_id)
    return note

def add_note(note, **kwargs):
    """
    Add a new note
    """
    note_i = Note()
    note_i.ref_key = note.ref_key

    note_i.set_ref(note.ref_key, note.ref_id)

    note_i.value = note.value

    note_i.created_by = kwargs.get('user_id')

    db.DBSession.add(note_i)
    db.DBSession.flush()

    return note_i

def update_note(note, **kwargs):
    """
    Update a note
    """
    note_i = _get_note(note.id)

    if note.ref_key != note_i.ref_key:
        raise HydraError("Cannot convert a %s note to a %s note. Please create a new note instead."%(note_i.ref_key, note.ref_key))

    note_i.set_ref(note.ref_key, note.ref_id)

    note_i.value = note.value

    db.DBSession.flush()

    return note_i

def purge_note(note_id, **kwargs):
    """
    Remove a note from the DB permenantly
    """
    note_i = _get_note(note_id)

    db.DBSession.delete(note_i)

    db.DBSession.flush()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from __future__ import division

from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import func
from sqlalchemy.orm import load_only

from .. import db

from ..util.dataset_util import array_dim
from ..util.dataset_util import arr_to_vector
from ..util.dataset_util import vector_to_arr
from ..db.model import Dataset, Unit, Dimension
from .objects import JSONObject
from ..exceptions import HydraError, ResourceNotFoundError, ValidationError

from ..util.permissions import required_perms

import numpy
import logging
log = logging.getLogger(__name__)



"""
+-------------------------------+
| DIMENSION FUNCTIONS - VARIOUS |
+-------------------------------+
"""

def exists_dimension(dimension_name,**kwargs):
    """
        Given a dimension returns True if it exists, False otherwise
    """
    try:
        dimension = db.DBSession.query(Dimension).filter(Dimension.name==dimension_name).one()
        # At this point the dimension exists
        return True
    except NoResultFound:
        # The dimension does not exist
        raise False

def is_global_dimension(dimension_id,**kwargs):
    """
        Returns True if the dimension is Global, False is it is assigned to a project
    """
    dimension = get_dimension(dimension_id)
    return (dimension.project_id is None)


"""
+--------------------------+
| UNIT FUNCTIONS - VARIOUS |
+--------------------------+
"""

def _parse_unit(measure_or_unit_abbreviation):
    """
        Helper function that extracts constant factors from unit specifications.
        This allows to specify units similar to this: 10^6 m^3.
        Return a couple (unit, factor)
    """
    try:
        float(measure_or_unit_abbreviation[0])
        # The measure contains the values and the unit_abbreviation
        factor, unit_abbreviation = measure_or_unit_abbreviation.split(' ', 1)
        return unit_abbreviation, float(factor)
    except ValueError:
        # The measure just contains the unit_abbreviation
        return measure_or_unit_abbreviation, 1.0

def is_global_unit(unit_id,**kwargs):
    """
        Returns True if the Unit is Global, False is it is assigned to a project
        'unit' is a Unit object
    """
    unit_data = get_unit(unit_id)
    return (unit_data.project_id is None)

def convert_units(values, source_measure_or_unit_abbreviation, target_measure_or_unit_abbreviation,**kwargs):
    """
        Convert a value from one unit to another one.

        Example::

            >>> cli = PluginLib.connect()
            >>> cli.service.convert_units(20.0, 'm', 'km')
            0.02
        Parameters:
            values: single measure or an array of measures
            source_measure_or_unit_abbreviation: A measure in the source unit, or just the abbreviation of the source unit, from which convert the provided measure value/values
            target_measure_or_unit_abbreviation: A measure in the target unit, or just the abbreviation of the target unit, into which convert the provided measure value/values

        Returns:
            Always a list
    """
    if numpy.isscalar(values):
        # If it is a scalar, converts to an array
        values = [values]
    float_values = [float(value) for value in values]
    values_to_return = convert(float_values, source_measure_or_unit_abbreviation, target_measure_or_unit_abbreviation)

    return values_to_return

def convert(values, source_measure_or_unit_abbreviation, target_measure_or_unit_abbreviation):
    """
        Convert a value or a list of values from an unit to another one.
        The two units must represent the same physical dimension.
    """

    source_dimension = get_dimension_by_unit_measure_or_abbreviation(source_measure_or_unit_abbreviation)
    target_dimension = get_dimension_by_unit_measure_or_abbreviation(target_measure_or_unit_abbreviation)

    if source_dimension == target_dimension:
        source=JSONObject({})
        target=JSONObject({})
        source.unit_abbreviation, source.factor = _parse_unit(source_measure_or_unit_abbreviation)
        target.unit_abbreviation, target.factor = _parse_unit(target_measure_or_unit_abbreviation)

        source.unit_data = get_unit_by_abbreviation(source.unit_abbreviation)
        target.unit_data = get_unit_by_abbreviation(target.unit_abbreviation)

        source.conv_factor = JSONObject({'lf': source.unit_data.lf, 'cf': source.unit_data.cf})
        target.conv_factor = JSONObject({'lf': target.unit_data.lf, 'cf': target.unit_data.cf})

        if isinstance(values, float):
            # If values is a float => returns a float
            return (source.conv_factor.lf / target.conv_factor.lf * (source.factor * values)
                    + (source.conv_factor.cf - target.conv_factor.cf)
                    / target.conv_factor.lf) / target.factor
        elif isinstance(values, list):
            # If values is a list of floats => returns a list of floats
            return [(source.conv_factor.lf / target.conv_factor.lf * (source.factor * value)
                    + (source.conv_factor.cf - target.conv_factor.cf)
                    / target.conv_factor.lf) / target.factor for value in values]
    else:
        raise HydraError("Unit conversion: dimensions are not consistent.")


"""
+---------------------------+
| DIMENSION FUNCTIONS - GET |
+---------------------------+
"""
def get_empty_dimension(**kwargs):
    """
        Returns a dimension object initialized with empty values
    """
    dimension = JSONObject(Dimension())
    dimension.id = None
    dimension.name = ''
    dimension.description = ''
    dimension.project_id = None
    dimension.units = []
    return dimension



def get_dimension(dimension_id, do_accept_dimension_id_none=False,**kwargs):
    """
        Given a dimension id returns all its data
    """
    if do_accept_dimension_id_none == True and dimension_id is None:
        # In this special case, the method returns a dimension with id None
        return get_empty_dimension()

    try:
        dimension = db.DBSession.query(Dimension).filter(Dimension.id==dimension_id).one()

        #lazy load units
        dimension.units

        return JSONObject(dimension)
    except NoResultFound:
        # The dimension does not exist
        raise ResourceNotFoundError("Dimension %s not found"%(dimension_id))


def get_dimensions(**kwargs):
    """
        Returns a list of objects describing all the dimensions with all the units.
    """
    dimensions_list = db.DBSession.query(Dimension).options(load_only(Dimension.id)).all()
    return_list = []
    for dimension in dimensions_list:
        return_list.append(get_dimension(dimension.id))


    return return_list


def get_dimension_by_name(dimension_name,**kwargs):
    """
        Given a dimension name returns all its data. Used in convert functions
    """
    try:
        if dimension_name is None:
            dimension_name = ''
        dimension = db.DBSession.query(Dimension).filter(func.lower(Dimension.name)==func.lower(dimension_name.strip())).one()

        return get_dimension(dimension.id)

    except NoResultFound:
        # The dimension does not exist
        raise ResourceNotFoundError("Dimension %s not found"%(dimension_name))


"""
+----------------------+
| UNIT FUNCTIONS - GET |
+----------------------+
"""


def get_unit(unit_id, **kwargs):
    """
        Returns a single unit
    """
    try:
        unit = db.DBSession.query(Unit).filter(Unit.id==unit_id).one()
        return JSONObject(unit)
    except NoResultFound:
        # The dimension does not exist
        raise ResourceNotFoundError("Unit %s not found"%(unit_id))

def get_units(**kwargs):
    """
        Returns all the units
    """
    units_list = db.DBSession.query(Unit).all()
    units = []
    for unit in units_list:
        new_unit = JSONObject(unit)
        units.append(new_unit)

    return units

def get_dimension_by_unit_measure_or_abbreviation(measure_or_unit_abbreviation,**kwargs):
    """
        Return the physical dimension a given unit abbreviation of a measure, or the measure itself, refers to.
        The search key is the abbreviation or the full measure
    """

    unit_abbreviation, factor = _parse_unit(measure_or_unit_abbreviation)

    units = db.DBSession.query(Unit).filter(Unit.abbreviation==unit_abbreviation).all()

    if len(units) == 0:
        raise HydraError('Unit %s not found.'%(unit_abbreviation))
    elif len(units) > 1:
        raise HydraError('Unit %s has multiple dimensions not found.'%(unit_abbreviation))
    else:
        dimension = db.DBSession.query(Dimension).filter(Dimension.id==units[0].dimension_id).one()
        return dimension

def get_dimension_by_unit_id(unit_id, do_accept_unit_id_none=False, **kwargs):
    """
        Return the physical dimension a given unit id refers to.
        if do_accept_unit_id_none is False, it raises an exception if unit_id is not valid or None
        if do_accept_unit_id_none is True, and unit_id is None, the function returns a Dimension with id None (unit_id can be none in some cases)
    """
    if do_accept_unit_id_none == True and unit_id is None:
        # In this special case, the method returns a dimension with id None
        return get_empty_dimension()

    try:
        dimension = db.DBSession.query(Dimension).join(Unit).filter(Unit.id==unit_id).filter().one()
        return get_dimension(dimension.id)
    except NoResultFound:
        # The dimension does not exist
        raise ResourceNotFoundError("Unit %s not found"%(unit_id))


def get_unit_by_abbreviation(unit_abbreviation, **kwargs):
    """
        Returns a single unit by abbreviation. Used as utility function to resolve string to id
    """
    try:
        if unit_abbreviation is None:
            unit_abbreviation = ''
        unit_i = db.DBSession.query(Unit).filter(Unit.abbreviation==unit_abbreviation.strip()).one()
        return JSONObject(unit_i)
    except NoResultFound:
        # The dimension does not exist
        raise ResourceNotFoundError("Unit '%s' not found"%(unit_abbreviation))

"""
+---------------------------------------+
| DIMENSION FUNCTIONS - ADD - DEL - UPD |
+---------------------------------------+
"""

@required_perms("add_dimension")
def add_dimension(dimension,**kwargs):
    """
        Add the dimension defined into the object "dimension" to the DB
        If dimension["project_id"] is None it means that the dimension is global, otherwise is property of a project
        If the dimension exists emits an exception
    """

    db_dimension = db.DBSession.query(Dimension).filter(Dimension.name==dimension['name']).first()

    if db_dimension is not None:
        raise HydraError(f"A dimension with name {dimension['name']} "
                         f"already exists with ID {db_dimension.id}")

    new_dimension = Dimension()
    new_dimension.name = dimension["name"]

    if "description" in dimension and dimension["description"] is not None:
        new_dimension.description = dimension["description"]
    if "project_id" in dimension and dimension["project_id"] is not None:
        new_dimension.project_id = dimension["project_id"]

    # Save on DB
    db.DBSession.add(new_dimension)
    db.DBSession.flush()

    # Load all the record
    db_dimension = db.DBSession.query(Dimension).filter(Dimension.id==new_dimension.id).one()
    #lazy load units
    db_dimension.units

    return JSONObject(db_dimension)

@required_perms("update_dimension")
def update_dimension(dimension,**kwargs):
    """
        Update a dimension in the DB.
        Raises and exception if the dimension does not exist.
        The key is ALWAYS the name and the name itself is not modificable
    """
    db_dimension = None
    dimension = JSONObject(dimension)
    try:
        db_dimension = db.DBSession.query(Dimension).filter(Dimension.id==dimension.id).filter().one()
        #laxy load units
        db_dimension.units
        if "description" in dimension and dimension["description"] is not None:
            db_dimension.description = dimension["description"]
        if "project_id" in dimension and dimension["project_id"] is not None and dimension["project_id"] != "" and dimension["project_id"].isdigit():
            db_dimension.project_id = dimension["project_id"]
    except NoResultFound:
        raise ResourceNotFoundError("Dimension (ID=%s) does not exist"%(dimension.id))


    db.DBSession.flush()
    return JSONObject(db_dimension)

@required_perms("delete_dimension")
def delete_dimension(dimension_id,**kwargs):
    """
        Delete a dimension from the DB. Raises and exception if the dimension does not exist
    """
    try:
        dimension = db.DBSession.query(Dimension).filter(Dimension.id==dimension_id).one()

        db.DBSession.query(Unit).filter(Unit.dimension_id==dimension.id).delete()

        db.DBSession.delete(dimension)
        db.DBSession.flush()
        return True
    except NoResultFound:
        raise ResourceNotFoundError("Dimension (dimension_id=%s) does not exist"%(dimension_id))


@required_perms("add_dimension")
def bulk_add_dimensions(dimension_list, **kwargs):
    """
        Save all the dimensions contained in the passed list.
    """
    added_dimensions = []
    for dimension in dimension_list:
        added_dimensions.append(add_dimension(dimension, **kwargs))

    return JSONObject({"dimensions": added_dimensions})

"""
+----------------------------------+
| UNIT FUNCTIONS - ADD - DEL - UPD |
+----------------------------------+
"""

@required_perms("add_unit")
def add_unit(unit,**kwargs):
    """
        Add the unit defined into the object "unit" to the DB
        If unit["project_id"] is None it means that the unit is global, otherwise is property of a project
        If the unit exists emits an exception


        A minimal example:

        .. code-block:: python

            new_unit = dict(

                name      = 'Teaspoons per second',
                abbreviation = 'tsp s^-1',
                cf        = 0,               # Constant conversion factor
                lf        = 1.47867648e-05,  # Linear conversion factor
                dimension_id = 2,
                description  = 'A flow of one teaspoon per second.',
            )
            add_unit(new_unit)


    """

    new_unit = Unit()
    new_unit.dimension_id   = unit["dimension_id"]
    new_unit.name           = unit['name']

    # Needed to uniform abbr to abbreviation
    new_unit.abbreviation = unit['abbreviation']

    # Needed to uniform into to description
    new_unit.description = unit.get('description')

    new_unit.lf             = unit['lf']
    new_unit.cf             = unit['cf']

    if ('project_id' in unit) and (unit['project_id'] is not None):
        # Adding dimension to the "user" dimensions list
        new_unit.project_id = unit['project_id']

    # Save on DB
    db.DBSession.add(new_unit)
    db.DBSession.flush()

    return JSONObject(new_unit)

@required_perms("add_unit")
def bulk_add_units(unit_list, **kwargs):
    """
        Save all the units contained in the passed list, with the name of their dimension.
    """
    # for unit in unit_list:
    #     add_unit(unit, **kwargs)

    added_units = []
    for unit in unit_list:
        added_units.append(add_unit(unit, **kwargs))

    return JSONObject({"units": added_units})




@required_perms("delete_unit")
def delete_unit(unit_id, **kwargs):
    """
        Delete a unit from the DB.
        Raises and exception if the unit does not exist
    """

    try:
        db_unit = db.DBSession.query(Unit).filter(Unit.id==unit_id).one()

        db.DBSession.delete(db_unit)
        db.DBSession.flush()
        return True
    except NoResultFound:
        raise ResourceNotFoundError("Unit (ID=%s) does not exist"%(unit_id))


@required_perms("update_unit")
def update_unit(unit, **kwargs):
    """
        Update a unit in the DB.
        Raises and exception if the unit does not exist
    """
    try:

        db_unit = db.DBSession.query(Unit).join(Dimension).filter(Unit.id==unit["id"]).filter().one()

        db_unit.name = unit["name"]

        # Needed to uniform into to description
        db_unit.abbreviation = unit.abbreviation
        db_unit.description = unit.description

        db_unit.lf = unit["lf"]
        db_unit.cf = unit["cf"]
        if "project_id" in unit and unit['project_id'] is not None and unit['project_id'] != "":
            db_unit.project_id = unit["project_id"]
    except NoResultFound:
        raise ResourceNotFoundError("Unit (ID=%s) does not exist"%(unit["id"]))


    db.DBSession.flush()
    return JSONObject(db_unit)



"""
+-----------------+
| OTHER FUNCTIONS |
+-----------------+
"""
def check_consistency(measure_or_unit_abbreviation, dimension,**kwargs):
    """
        Check whether a specified unit is consistent with the physical
        dimension asked for by the attribute or the dataset.
    """
    dim = get_dimension_by_unit_measure_or_abbreviation(measure_or_unit_abbreviation)
    return dim.name == dimension

def check_unit_matches_dimension(unit_id, dimension_id,**kwargs):
    """
        Check whether a specified unit is part of the specified dimension.
        args:
            unit_id (int): The ID of the unit to compare with the dimension
            dimension_id (int): The ID of the dimension to check
        throws
            hydra_base.ValidationError when the unit_id does not match the supplied dimension
        returns:
            None
    """
    #ensure the unit exists
    unit_i = db.DBSession.query(Unit).filter(Unit.id==unit_id).one()

    if dimension_id is not None:
        #ensure the dimension exists
        dimension_i = db.DBSession.query(Dimension).filter(Dimension.id==dimension_id).one()
        dimension_name = dimension_i.name
    else:
        dimension_name = None

    if unit_i.dimension.id != dimension_id:
        raise ValidationError(f"Unit {unit_i.name} has a dimension of {unit_i.dimension.name}, not {dimension_name}")


"""
+-------------------+
| DATASET functions |
+-------------------+
"""


def convert_dataset(dataset_id, target_unit_abbreviation,**kwargs):
    """
        Convert a whole dataset (specified by 'dataset_id') to new unit ('target_unit_abbreviation').
        Conversion ALWAYS creates a NEW dataset, so function returns the dataset ID of new dataset.
    """

    ds_i = db.DBSession.query(Dataset).filter(Dataset.id==dataset_id).one()

    dataset_type = ds_i.type

    dsval = ds_i.get_val()
    source_unit_abbreviation = get_unit(ds_i.unit_id).abbreviation

    if source_unit_abbreviation is not None:
        if dataset_type == 'scalar':
            new_val = convert(float(dsval), source_unit_abbreviation, target_unit_abbreviation)
        elif dataset_type == 'array':
            dim = array_dim(dsval)
            vecdata = arr_to_vector(dsval)
            newvec = convert(vecdata, source_unit_abbreviation, target_unit_abbreviation)
            new_val = vector_to_arr(newvec, dim)
        elif dataset_type == 'timeseries':
            new_val = []
            for ts_time, ts_val in dsval.items():
                dim = array_dim(ts_val)
                vecdata = arr_to_vector(ts_val)
                newvec = convert(vecdata, source_unit_abbreviation, target_unit_abbreviation)
                newarr = vector_to_arr(newvec, dim)
                new_val.append(ts_time, newarr)
        elif dataset_type == 'descriptor':
            raise HydraError('Cannot convert descriptor.')

        new_dataset = Dataset()
        new_dataset.type   = ds_i.type
        new_dataset.value  = str(new_val)
        new_dataset.name   = ds_i.name

        new_dataset.unit_id   = get_unit_by_abbreviation(target_unit_abbreviation).id
        new_dataset.hidden = 'N'
        new_dataset.set_metadata(ds_i.get_metadata_as_dict())
        new_dataset.set_hash()

        existing_ds = db.DBSession.query(Dataset).filter(Dataset.hash==new_dataset.hash).first()

        if existing_ds is not None:
            db.DBSession.expunge_all()
            return existing_ds.id

        db.DBSession.add(new_dataset)
        db.DBSession.flush()

        return new_dataset.id

    else:
        raise HydraError('Dataset has no units.')


"""
+--------------------+
| RESOURCE functions |
+--------------------+
"""

def validate_resource_attributes(resource, attributes, template, check_unit=True, exact_match=False,**kwargs):
    """
        Validate that the resource provided matches the template.
        Only passes if the resource contains ONLY the attributes specified
        in the template.

        The template should take the form of a dictionary, as should the
        resources.

        *check_unit*: Makes sure that if a unit is specified in the template, it
                      is the same in the data
        *exact_match*: Ensures that all the attributes in the template are in
                       the data also. By default this is false, meaning a subset
                       of the template attributes may be specified in the data.
                       An attribute specified in the data *must* be defined in
                       the template.

        @returns a list of error messages. An empty list indicates no
        errors were found.
    """
    errors = []
    #is it a node or link?
    res_type = 'GROUP'
    if resource.get('x') is not None:
        res_type = 'NODE'
    elif resource.get('node_1_id') is not None:
        res_type = 'LINK'
    elif resource.get('nodes') is not None:
        res_type = 'NETWORK'

    #Find all the node/link/network definitions in the template
    tmpl_res = template['resources'][res_type]

    #the user specified type of the resource
    res_user_type = resource.get('type')

    #Check the user specified type is in the template
    if res_user_type is None:
        errors.append("No type specified on resource %s"%(resource['name']))

    elif tmpl_res.get(res_user_type) is None:
        errors.append("Resource %s is defined as having type %s but "
                      "this type is not specified in the template."%
                      (resource['name'], res_user_type))

    #It is in the template. Now check all the attributes are correct.
    tmpl_attrs = tmpl_res.get(res_user_type)['attributes']

    attrs = {}
    for a in attributes.values():
        attrs[a['id']] = a

    for a in tmpl_attrs.values():
        if a.get('id') is not None:
            attrs[a['id']] = {'name':a['name'], 'unit':a.get('unit'), 'dimen':a.get('dimension')}

    if exact_match is True:
        #Check that all the attributes in the template are in the data.
        #get all the attribute names from the template
        tmpl_attr_names = set(tmpl_attrs.keys())
        #get all the attribute names from the data for this resource
        resource_attr_names = []
        for ra in resource['attributes']:
            attr_name = attrs[ra['attr_id']]['name']
            resource_attr_names.append(attr_name)
        resource_attr_names = set(resource_attr_names)

        #Compare the two lists to ensure they are the same (using sets is easier)
        in_tmpl_not_in_resource = tmpl_attr_names - resource_attr_names
        in_resource_not_in_tmpl = resource_attr_names - tmpl_attr_names

        if len(in_tmpl_not_in_resource) > 0:
            errors.append("Template has defined attributes %s for type %s but they are not"
                            " specified in the Data."%(','.join(in_tmpl_not_in_resource),
                                                    res_user_type ))

        if len(in_resource_not_in_tmpl) > 0:
            errors.append("Resource %s (type %s) has defined attributes %s but this is not"
                            " specified in the Template."%(resource['name'],
                                                        res_user_type,
                                                        ','.join(in_resource_not_in_tmpl)))

    #Check that each of the attributes specified on the resource are valid.
    for res_attr in resource['attributes']:

        attr = attrs.get(res_attr['attr_id'])

        if attr is None:
            errors.append("An attribute mismatch has occurred. Attr %s is not "
                          "defined in the data but is present on resource %s"
                          %(res_attr['attr_id'], resource['name']))
            continue

        #If an attribute is not specified in the template, then throw an error
        if tmpl_attrs.get(attr['name']) is None:
            errors.append("Resource %s has defined attribute %s but this is not"
                          " specified in the Template."%(resource['name'], attr['name']))
        else:
            #If the dimensions or units don't match, throw an error

            tmpl_attr = tmpl_attrs[attr['name']]

            if tmpl_attr.get('data_type') is not None:
                if res_attr.get('type') is not None:
                    if tmpl_attr.get('data_type') != res_attr.get('type'):
                        errors.append("Error in data. Template says that %s on %s is a %s, but data suggests it is a %s"%
                            (attr['name'], resource['name'], tmpl_attr.get('data_type'), res_attr.get('type')))

            attr_dimension = 'dimensionless' if attr.get('dimension') is None else attr.get('dimension')
            tmpl_attr_dimension = 'dimensionless' if tmpl_attr.get('dimension') is None else tmpl_attr.get('dimension')

            if attr_dimension.lower() != tmpl_attr_dimension.lower():
                errors.append("Dimension mismatch on resource %s for attribute %s"
                              " (template says %s on type %s, data says %s)"%
                              (resource['name'], attr.get('name'),
                               tmpl_attr.get('dimension'), res_user_type, attr_dimension))

            if check_unit is True:
                if tmpl_attr.get('unit') is not None:
                    if attr.get('unit') != tmpl_attr.get('unit'):
                        errors.append("Unit mismatch for resource %s with unit %s "
                                      "(template says %s for type %s)"
                                      %(resource['name'], attr.get('unit'),
                                        tmpl_attr.get('unit'), res_user_type))

    return errors

"""
+-------------------+
| Utility functions |
+-------------------+
"""



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
import sys
import logging
import json
import pytest
import datetime

from hydra_base.exceptions import HydraError
from hydra_base.lib.objects import JSONObject
import hydra_base as hb

log = logging.getLogger(__name__)

"""
+----------------------------+
| NEEDED for testing dataset |
+----------------------------+
"""

def arr_to_vector(arr):
    """Reshape a multidimensional array to a vector.
    """
    dim = array_dim(arr)
    tmp_arr = []
    for n in range(len(dim) - 1):
        for inner in arr:
            for i in inner:
                tmp_arr.append(i)
        arr = tmp_arr
        tmp_arr = []
    return arr

def array_dim(arr):
    """Return the size of a multidimansional array.
    """
    dim = []
    while True:
        try:
            dim.append(len(arr))
            arr = arr[0]
        except TypeError:
            return dim



class TestUnits():
    """
        Test for working with units.
    """
    def test_get_empty_dimension(self, client):
        dimension = client.get_empty_dimension()
        assert dimension.id is None, "get_empty_dimension didn't work as expected"


    def test_get_dimension(self, client):
        # Returns a dimension searching by ID. The result contains all the
        # dimension data plus all the units of the dimension
        # 1 is the id of length
        dimension = client.get_dimension(1)

        assert dimension is not None, \
            "Could not get a dimensions by id."
        assert dimension.name == "Length", \
            "Could not get the dimension 'Length' by ID 1."
        assert len(dimension.units) > 0, \
            "Could not get the dimension units for Dimension ID 1."

        with pytest.raises(HydraError):
            # It must raise an exception
            dimension = client.get_dimension(None, do_accept_dimension_id_none=False)

        dimension = client.get_dimension(None, do_accept_dimension_id_none=True)
        assert dimension is not None, \
            "Could not get a dimension by None."
        assert dimension.id is None, \
            "Could not get a dimension by None."


    def test_get_dimensions(self, client):
        # Returns an array of dimensions. Every item contains the
        # dimension data plus the units list
        dimension_list = client.get_dimensions()

        assert dimension_list is not None and len(dimension_list) != 0, \
            "Could not get the full list of dimensions."
        assert dimension_list[0].id is not None, \
            "The first element of the list has no ID.  An ID is expected on a dimension"

    def test_get_dimension_by_name(self, client):
        # Returns a dimension searching by ID. The result contains all the
        # dimension data plus all the units of the dimension
        # 1 is the id of length
        dimension = client.get_dimension_by_name("Length")

        assert dimension is not None, \
            "Could not get a dimensions by name."
        assert dimension.id == 1, \
            "Could not get the dimension 1 by aname 'Length'."
        assert len(dimension.units) > 0, \
            "Could not get the dimension units for Dimension ID 1."

    def test_get_unit(self, client):
        # Returns a dimension searching by ID. The result contains all the
        # dimension data plus all the units of the dimension
        # 1 is the id of length
        unit = client.get_unit(2)

        assert unit is not None, \
            "Could not get a unit by id."
        assert unit.abbreviation == "AU", \
            "Could not get the unit 'AU (astronomical unit)' by ID 2."

    def test_get_unit_by_abbreviation(self, client):
        # Returns a dimension searching by ID. The result contains all the
        # dimension data plus all the units of the dimension
        # 1 is the id of length
        unit = client.get_unit_by_abbreviation("AU")

        assert unit is not None, \
            "Could not get a unit by abbreviation."
        assert unit.id == 2, \
            "Could not get the unit '(astronomical unit)' with ID 2 by abbreviation 'AU'."

    def test_get_units(self, client):
        # Returns an array of dimensions. Every item contains the dimension data plus the units list
        units_list = client.get_units()

        assert units_list is not None and len(units_list) != 0, \
            "Could not get the full list of units."


    def test_get_dimension_by_unit_measure_or_abbreviation(self, client):

        testdim = 'Length'
        testunit = 'km'
        resultdim = client.get_dimension_by_unit_measure_or_abbreviation(testunit)

        assert testdim == resultdim.name, \
            "Getting dimension for 'km' didn't work."

        with pytest.raises(HydraError):
            dimension = client.get_dimension_by_unit_measure_or_abbreviation('not-existing-unit')

    def test_get_dimension_by_unit_id(self, client):
        # Returns a dimension searching by ID. The result contains all the
        # dimension data plus all the units of the dimension
        # 1 is the id of length

        # Referring dimension
        testdim = JSONObject({'name':'Length', 'id': 1})
        # Units list of the dimension
        units = client.get_dimension(testdim.id).units
        # The first unit id
        test_unit_id = units[0].id
        # The dimension relative to the first unit id
        dimension = client.get_dimension_by_unit_id(test_unit_id)

        assert dimension is not None, \
            "Could not get a dimension by unit id."
        assert dimension.id == testdim.id, \
            "Could not get the dimension '{}' by unit id {}.".format(testdim.name, test_unit_id)

        with pytest.raises(HydraError):
            # It must raise an exception
            dimension = client.get_dimension_by_unit_id(None, do_accept_unit_id_none=False)

        dimension = client.get_dimension_by_unit_id(None, do_accept_unit_id_none=True)
        assert dimension is not None, \
            "Could not get a dimension by None."
        assert dimension.id is None, \
            "Could not get a dimension by None."


    """
        Manipulate DIMENSIONS
    """

    def test_add_dimension(self, client):

        # Try to add an existing dimension
        testdim = {'name': 'Length'}
        with pytest.raises(Exception) as excinfo:
            client.add_dimension(testdim)


        # IN this way we can test that the exception is specifically what we
        # want and not other exceptions type
        assert "already exists" in str(excinfo.value)

        # Add a new dimension
        testdim = {'name':'Electric current'}
        client.add_dimension(testdim)

        dimension_list = list(client.get_dimensions())

        assert testdim['name'] in [d.name for d in dimension_list]


    def test_update_dimension(self, client):
        # Updating existing dimension
        testdim = {
            'id': 1,
            'name':'Length',
            'description': 'New description'
        }
        modified_dim = client.update_dimension(testdim)

        assert modified_dim.description == testdim["description"], \
                "Updating a dimension didn't work"


    def test_delete_dimension(self, client):
        # Add a new dimension and delete it

        # Test adding the object and deleting it
        testdim = {'name':f'Dimension {datetime.datetime.now()}'}
        new_dimension = client.add_dimension(testdim)

        old_dimension_list = list(client.get_dimensions())

        client.delete_dimension(new_dimension.id)

        new_dimension_list = list(client.get_dimensions())

        assert len(list(filter(lambda x: x["name"] == testdim["name"], old_dimension_list))) > 0 and \
               len(list(filter(lambda x: x["name"] == testdim["name"], new_dimension_list))) == 0,\
            "Deleting a dimension didn't work as expected."

    """
        Manipulate UNITS
    """
    def test_add_unit(self, client):
        # Add a new unit to an existing static dimension
        new_unit = JSONObject({})
        new_unit.name = 'Teaspoons per second'
        new_unit.abbreviation = 'tsp s^-1'
        new_unit.cf = 0               # Constant conversion factor
        new_unit.lf = 1.47867648e-05  # Linear conversion factor

        new_unit.dimension_id = client.get_dimension_by_name('Volumetric flow rate').id
        new_unit.info = 'A flow of one tablespoon per second.'
        client.add_unit(new_unit)

        dimension_loaded = client.get_dimension(new_unit.dimension_id)

        assert len(list(filter(lambda x: x["name"] == new_unit["name"], dimension_loaded.units))) > 0 , \
            "Adding new unit didn't work."


        # Add a new unit to a custom dimension
        testdim = {'name':'Test dimension'}
        new_dimension = client.add_dimension(testdim)

        testunit = JSONObject({})
        testunit.name = 'Test'
        testunit.abbreviation = 'ttt'
        testunit.cf = 21
        testunit.lf = 42

        testunit.dimension_id = new_dimension.id

        new_unit = client.add_unit(testunit)

        new_dimension_loaded = client.get_dimension(new_dimension.id)

        unitlist = list(new_dimension_loaded.units)

        assert len(unitlist) == 1, \
            "Adding a new unit didn't work as expected"

        assert unitlist[0]["name"] == 'Test', \
            "Adding a new unit didn't work as expected"

        client.delete_dimension(new_dimension.id)



    def test_update_unit(self, client):
        # Add a new unit to a new dimension

        testdim = {'name':'Test dimension'}
        new_dimension = client.add_dimension(testdim)

        testunit = JSONObject({})
        testunit.name = 'Test'
        testunit.abbreviation = 'ttt'
        testunit.cf = 21
        testunit.lf = 42
        testunit.dimension_id = new_dimension.id
        new_unit = client.add_unit(testunit)

        # Update it
        new_unit.cf = 0
        new_unit_modified = client.update_unit(new_unit)

        unitlist = list(client.get_dimension(new_dimension.id).units)

        assert len(unitlist) > 0 and int(unitlist[0]['cf']) == 0, \
            "Updating unit didn't work correctly."

        client.delete_dimension(new_dimension.id)

    def test_delete_unit(self, client):
        # Add a new unit to a new dimension

        testdim = {'name':'Test dimension'}
        new_dimension = client.add_dimension(testdim)

        testunit = JSONObject({})
        testunit.name = 'Test'
        testunit.abbreviation = 'ttt'
        testunit.cf = 21
        testunit.lf = 42
        testunit.dimension_id = new_dimension.id
        new_unit = client.add_unit(testunit)

        # Check if the unit has been added
        unitlist = list(client.get_dimension(new_dimension.id).units)


        assert len(unitlist) > 0 and unitlist[0]['abbreviation'] == testunit.abbreviation, \
            "The adding has not worked properly"

        result = client.delete_unit(new_unit.id)

        unitlist = list(client.get_dimension(new_dimension.id).units)

        assert len(unitlist) == 0, \
            "Deleting unit didn't work correctly."

        client.delete_dimension(new_dimension.id)

    def test_convert_units(self, client):

        result = client.convert_units(20, 'm', 'km')
        assert result == [0.02], \
            "Converting metres to kilometres didn't work."

        result = client.convert_units([20., 30., 40.], 'm', 'km')
        assert result == [0.02, 0.03, 0.04],  \
            "Unit conversion of array didn't work."

        result = client.convert_units(20, '2e6 m^3', 'hm^3')
        assert result == [40], "Conversion with factor didn't work correctly."



    def test_check_consistency(self, client):
        result1 = client.check_consistency('m^3', 'Volume')
        result2 = client.check_consistency('m', 'Volume')
        assert result1 is True, \
            "Unit consistency check didn't work."
        assert result2 is False, \
            "Unit consistency check didn't work."



    def test_is_global_dimension(self, client):
        result = client.is_global_dimension(1)
        assert result is True, \
            "Is global dimension check didn't work."

    def test_is_global_unit(self, client):
        result = client.is_global_unit(1)
        assert result is True, \
            "Is global unit check didn't work."

    def test_convert_dataset(self, client):
        project = client.testutils.create_project()

        network = client.testutils.create_network_with_data(num_nodes=2, project_id=project.id)

        resourcescenarios = \
            network.scenarios[0].resourcescenarios

        # Select the first array (should have untis 'bar') and convert it
        for res_scen in resourcescenarios:
            if res_scen.dataset.type == 'array':
                dataset_id = res_scen.dataset.id
                old_val = res_scen.dataset.value
                break

        newid = client.convert_dataset(dataset_id, 'mmHg')

        assert newid is not None
        assert newid != dataset_id, "Converting dataset not completed."

        new_dataset = client.get_dataset(newid)
        new_val = new_dataset.value

        new_val = arr_to_vector(json.loads(new_val))
        old_val = arr_to_vector(json.loads(old_val))

        old_val_conv = [i * 100000 / 133.322 for i in old_val]

        # Rounding is not exactly the same on the server, that's why we
        # calculate the sum.
        assert sum(new_val) - sum(old_val_conv) < 0.00001, \
            "Unit conversion did not work"

    def test_apply_unit_to_network_rs(self, client, network_with_data):

        network_id = network_with_data.id

        dimension_list = client.get_dimensions()

        rs_to_change = network_with_data.scenarios[0].resourcescenarios[0]

        #The ID of the attribute for which these changes will be made
        attr_id = rs_to_change.resourceattr.attr_id

        #identify a unit to change
        unit_to_change = rs_to_change.dataset.unit_id

        #find another valid unit to replace it with
        unit = client.get_unit(unit_to_change)
        dimension = client.get_dimension(unit.dimension_id)

        new_unit = dimension.units[0].id
        client.apply_unit_to_network_rs(network_id, new_unit, attr_id)

        #now try to apply an incompatible unit. Just go grab another dimension, ensuring it's not accidenally
        #the one we're using
        bad_dimension = dimension_list[0]
        if bad_dimension.id == dimension.id:
            bad_dimension == dimension_list[1]

        bad_unit = bad_dimension.units[0].id

        with pytest.raises(HydraError):
            client.apply_unit_to_network_rs(network_id, bad_unit, attr_id)

        updated_scenario = client.get_scenario(rs_to_change.scenario_id)

        #it's possible that the same unit spans multiple attributes in the original scenario
        #so the updated scenario can only be tested for datsets which are attached to the attribute
        old_unit_rs = [rs for rs in network_with_data.scenarios[0].resourcescenarios if rs.resourceattr.attr_id==attr_id]
        new_unit_rs = [rs for rs in updated_scenario.resourcescenarios if rs.resourceattr.attr_id==attr_id]

        #check that the attributes haven't changed
        assert len(old_unit_rs) == len(new_unit_rs)
        #check that all units have been updated
        assert set([rs.dataset.unit_id for rs in new_unit_rs]) == {new_unit}

        for rs in updated_scenario.resourcescenarios:
            if rs.resourceattr.attr_id == attr_id:
                assert rs.dataset.unit_id == new_unit



"""dataset_audit

Revision ID: d759891ddbf4
Revises: da2a4916c900
Create Date: 2019-11-14 13:44:19.028436

"""
from alembic import op
import sqlalchemy as sa

import datetime

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = 'd759891ddbf4'
down_revision = 'da2a4916c900'
branch_labels = None
depends_on = None


def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tScenario
        try:
            op.add_column('tDataset', sa.Column('updated_by', sa.Integer(), sa.ForeignKey('tUser.id'), nullable=True))
            op.add_column('tDataset', sa.Column('updated_at', sa.DateTime(), default=datetime.datetime.utcnow(), onupdate=datetime.datetime.utcnow(), nullable=True))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tDataset_new')
        except:
            log.info('tDataset_new does not exist')

        try:
            # ## tScenario
            op.create_table(
                'tDataset_new',
                sa.Column('id', sa.Integer(), primary_key=True, index=True, nullable=False),
                sa.Column('name', sa.String(200),  nullable=False),
                sa.Column('type', sa.String(60),  nullable=False),
                sa.Column('unit_id', sa.Integer(), sa.ForeignKey('tUnit.id'),  nullable=True),
                sa.Column('hash', sa.BIGINT(),  nullable=False, unique=True),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                sa.Column('updated_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                sa.Column('updated_at', sa.DateTime(), sa.ForeignKey('tUser.id')),
                sa.Column('hidden', sa.String(1),  nullable=False, server_default=sa.text(u"'N'")),
                sa.Column('value', sa.Text(),  nullable=True)
            )

            op.execute("insert into tDataset_new (id, name, type, unit_id, hash, cr_date, created_by, updated_by, updated_at, hidden, value) select id, name, type, unit_id, hash, cr_date, created_by, null, null, hidden, value from tDataset")

            op.rename_table('tDataset','tDataset_old')
            op.rename_table('tDataset_new', 'tDataset')
            op.drop_table('tDataset_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tScenario
        try:
            op.drop_column('tDataset',  'updated_at')
            op.drop_column('tDataset', 'updated_by')
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tDataset_new')
        except:
            log.info('tDataset_new does not exist')

        try:
            # ## tScenario
            op.create_table(
                'tDataset_new',
                sa.Column('id', sa.Integer(), primary_key=True, index=True, nullable=False),
                sa.Column('name', sa.String(200),  nullable=False),
                sa.Column('type', sa.String(60),  nullable=False),
                sa.Column('unit_id', sa.Integer(), sa.ForeignKey('tUnit.id'),  nullable=True),
                sa.Column('hash', sa.BIGINT(),  nullable=False, unique=True),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id')),
                sa.Column('hidden', sa.String(1),  nullable=False, server_default=sa.text(u"'N'")),
                sa.Column('value', sa.Text(),  nullable=True)
            )

            op.execute("insert into tDataset_new (id, name, type, unit_id, hash, cr_date, created_by, hidden, value) select id, name, type, unit_id, hash, cr_date, created_by, hidden, value from tDataset")

            op.rename_table('tDataset','tDataset_old')
            op.rename_table('tDataset_new', 'tDataset')
            op.drop_table('tDataset_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

from ..base import *

from ..template import Template, TemplateType
from ..ownership import NetworkOwner
from .resourceattr import ResourceAttr
from ..scenario import ResourceGroupItem
from ..attributes import Attr

from . import Link
from . import Node
from . import ResourceGroup
from .resource import Resource

__all__ = ['Network']



class Network(Base, Inspect, PermissionControlled, Resource):
    """
    """

    __tablename__ = 'tNetwork'
    __table_args__ = (
        UniqueConstraint('name', 'project_id', name="unique net name"),
    )
    ref_key = 'NETWORK'

    id = Column(Integer(), primary_key=True, nullable=False)
    name = Column(String(200), nullable=False)
    description = Column(String(1000))
    layout = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'), nullable=True)
    appdata = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'), nullable=True)
    project_id = Column(Integer(), ForeignKey('tProject.id'), nullable=False)
    status = Column(String(1), nullable=False, server_default=text(u"'A'"))
    cr_date = Column(TIMESTAMP(), nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))
    projection = Column(String(200))
    created_by = Column(Integer(), ForeignKey('tUser.id'), nullable=False)

    project = relationship('Project',
                           backref=backref("networks",
                                           order_by="asc(Network.cr_date)",
                                           cascade="all, delete-orphan"))

    _parents = ['tNode', 'tLink', 'tResourceGroup']
    _children = ['tProject']

    def is_owner(self, user_id):
        """
            Check whether this user is an owner of this project, either directly
            or by virtue of being an owner of a higher-up project
        """

        if self.check_read_permission(user_id, do_raise=False) is True:
            return True

        else:
            return self.project.is_owner(user_id)

    def get_name(self):
        return self.name

    def add_attribute(self, attr_id, attr_is_var='N'):
        res_attr = ResourceAttr()
        res_attr.attr_id = attr_id
        res_attr.attr_is_var = attr_is_var
        res_attr.ref_key = self.ref_key
        res_attr.network_id  = self.id
        self.attributes.append(res_attr)

        return res_attr

    def add_link(self, name, desc, layout, node_1, node_2):
        """
            Add a link to a network. Links are what effectively
            define the network topology, by associating two already
            existing nodes.
        """

        existing_link = get_session().query(Link).filter(Link.name==name, Link.network_id==self.id).first()
        if existing_link is not None:
            raise HydraError("A link with name %s is already in network %s"%(name, self.id))

        l = Link()
        l.name        = name
        l.description = desc
        l.layout           = json.dumps(layout) if layout is not None else None
        l.node_a           = node_1
        l.node_b           = node_2

        get_session().add(l)

        self.links.append(l)

        return l


    def add_node(self, name, desc, layout, node_x, node_y):
        """
            Add a node to a network.
        """
        existing_node = get_session().query(Node).filter(Node.name==name, Node.network_id==self.id).first()
        if existing_node is not None:
            raise HydraError("A node with name %s is already in network %s"%(name, self.id))

        node = Node()
        node.name        = name
        node.description = desc
        node.layout      = str(layout) if layout is not None else None
        node.x           = node_x
        node.y           = node_y

        #Do not call save here because it is likely that we may want
        #to bulk insert nodes, not one at a time.

        get_session().add(node)

        self.nodes.append(node)

        return node

    def add_group(self, name, desc, status):
        """
            Add a new group to a network.
        """

        existing_group = get_session().query(ResourceGroup).filter(ResourceGroup.name==name, ResourceGroup.network_id==self.id).first()
        if existing_group is not None:
            raise HydraError("A resource group with name %s is already in network %s"%(name, self.id))

        group_i             = ResourceGroup()
        group_i.name        = name
        group_i.description = desc
        group_i.status      = status

        get_session().add(group_i)

        self.resourcegroups.append(group_i)


        return group_i

    def set_owner(self, user_id, read='Y', write='Y', share='Y'):
        owner = None
        for o in self.owners:
            if str(user_id) == str(o.user_id):
                owner = o
                break
        else:
            owner = NetworkOwner()
            owner.network_id = self.id
            self.owners.append(owner)

        owner.user_id = int(user_id)
        owner.view  = read
        owner.edit  = write
        owner.share = share

        return owner

    def unset_owner(self, user_id):

        owner = None
        if str(user_id) == str(self.created_by):
            log.warning("Cannot unset %s as owner, as they created the network", user_id)
            return
        for o in self.owners:
            if user_id == o.user_id:
                owner = o
                get_session().delete(owner)
                break

    def check_read_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can read this network
        """

        can_read = super(Network, self).check_read_permission(user_id, do_raise=False, is_admin=is_admin)

        if can_read is True:
            return True

        can_read = self.project.check_read_permission(user_id, do_raise=False)

        if can_read is False and do_raise is True:
            raise PermissionError("Permission denied. User %s does not have read"
                         " access on network %s" %
                         (user_id, self.id))

        return can_read

    def check_write_permission(self, user_id, do_raise=True, is_admin=None):
        """
            Check whether this user can write this project
        """
        can_write = super(Network, self).check_write_permission(user_id, do_raise=False, is_admin=is_admin)

        if can_write is True:
            return True

        can_write = self.project.check_write_permission(user_id)

        if can_write is False and do_raise is True:
            raise PermissionError("Permission denied. User %s does not have edit"
                         " access on network %s" %
                         (user_id, self.id))

        return can_write

    def get_owners(self):
        """
            Get all the owners of a network, both those which are applied directly
            to this network, but also who have been granted access via a project
        """

        owners = [JSONObject(o) for o in self.owners]
        owner_ids = [o.user_id for o in owners]

        project_owners = list(filter(lambda x:x.user_id not in owner_ids, self.project.get_owners()))

        for po in project_owners:
            po.source = f'Inherited from: {po.project_name} (ID:{po.project_id})'

        return owners + project_owners

    def get_scoped_attributes(self, include_hierarchy=False, name_match=None, return_json=True):
        """
            Get all the attributes scoped to this project, and to all projects above
            it in the project hierarchy (including global attributes if requested)
            args:
                include_hierarchy (Bool): Include attribtues from projects higher up in the 
                    project hierarchy
        """

        scoped_attrs_qry = get_session().query(Attr).filter(Attr.network_id==self.id)

        if name_match is not None:
            name_match = name_match.lower()
            scoped_attrs_qry = scoped_attrs_qry.filter(
                func.lower(Attr.name).like(f'%{name_match}%'))

        scoped_attrs = scoped_attrs_qry.all()

        if include_hierarchy is True:
            scoped_attrs.extend(self.project.get_scoped_attributes(
                include_hierarchy=True, name_match=name_match))

        if return_json is True:
            
            scoped_attrs_j = [JSONObject(a) for a in scoped_attrs]
            #This is for convenience to avoid having to do extra calls to get the network name
            for a in scoped_attrs_j:
                a.network_name = self.name
            return scoped_attrs_j
        else:
            return scoped_attrs


"""scenario_id

Revision ID: 3dc489a7eac1
Revises: be206abd1412
Create Date: 2018-04-04 16:16:07.105230

"""
from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import mysql

# revision identifiers, used by Alembic.
revision = '3dc489a7eac1'
down_revision = 'be206abd1412'
branch_labels = None
depends_on = None

import logging
log = logging.getLogger(__name__)

def upgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tScenario

        try:
            op.alter_column('tScenario', 'scenario_id', new_column_name='id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tScenario', 'scenario_name', new_column_name='name', existing_type=sa.String(200))
            op.alter_column('tScenario', 'scenario_description', new_column_name='description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tScenario_new')
        except:
            log.info('tScenario_new does not exist')

        try:
            # ## tScenario
            op.create_table(
                'tScenario_new',
                sa.Column('id', sa.Integer(), primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('layout', sa.Text(1000)),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id'),  nullable=False),
                sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'), index=True),
                sa.Column('start_time', sa.String(60)),
                sa.Column('end_time', sa.String(60)),
                sa.Column('locked', sa.String(1),  nullable=False, server_default=sa.text(u"'N'")),
                sa.Column('time_step', sa.String(60)),
                sa.UniqueConstraint('network_id', 'name', name="unique scenario name"),
            )

            op.execute("insert into tScenario_new (id, name, description, layout, status, cr_date, created_by, network_id, start_time, end_time, time_step, locked) select scenario_id, scenario_name, scenario_description, layout, status, cr_date, created_by, network_id, start_time, end_time, time_step, locked from tScenario")

            op.rename_table('tScenario','tScenario_old')
            op.rename_table('tScenario_new', 'tScenario')
            op.drop_table('tScenario_old')

        except Exception as e:
            log.exception(e)

def downgrade():
    if op.get_bind().dialect.name == 'mysql':

        # ### tScenario

        try:
            op.alter_column('tScenario', 'id', new_column_name='scenario_id', existing_type=sa.Integer(), autoincrement=True, nullable=False)
            op.alter_column('tScenario', 'name', new_column_name='scenario_name', existing_type=sa.String(200))
            op.alter_column('tScenario', 'description', new_column_name='scenario_description', existing_type=sa.String(1000))
        except Exception as e:
            log.exception(e)

    else: ## sqlite

        try:
            op.drop_table('tScenario_new')
        except:
            log.info('tScenario_new does not exist')

        try:
            # ## tScenario
            op.create_table(
                'tScenario_new',
                sa.Column('id', sa.Integer(), primary_key=True),
                sa.Column('name', sa.Text(200), nullable=False),
                sa.Column('description', sa.Text(1000)),
                sa.Column('layout', sa.Text(1000)),
                sa.Column('status', sa.String(1),  nullable=False, server_default=sa.text(u"'A'")),
                sa.Column('cr_date', sa.TIMESTAMP(),  nullable=False, server_default=sa.text(u'CURRENT_TIMESTAMP')),
                sa.Column('created_by', sa.Integer(), sa.ForeignKey('tUser.id'),  nullable=False),
                sa.Column('network_id', sa.Integer(), sa.ForeignKey('tNetwork.id'), index=True),
                sa.Column('start_time', sa.String(60)),
                sa.Column('end_time', sa.String(60)),
                sa.Column('locked', sa.String(1),  nullable=False, server_default=sa.text(u"'N'")),
                sa.Column('time_step', sa.String(60)),
                sa.UniqueConstraint('network_id', 'scenario_name', name="unique scenario name"),
            )

            op.execute("insert into tScenario_new (scenario_id, scenario_name, scenario_description, layout, status, cr_date, created_by, network_id, start_time, end_time, time_step, locked) select id, name, description, layout, status, cr_date, created_by, network_id, start_time, end_time, time_step, locked from tScenario")

            op.rename_table('tScenario','tScenario_old')
            op.rename_table('tScenario_new', 'tScenario')
            op.drop_table('tScenario_old')

        except Exception as e:
            log.exception(e)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#

import sqlalchemy
from sqlalchemy.orm import scoped_session
from sqlalchemy import create_engine

#Import these as a test for foreign key checking in
from sqlalchemy import event, text
from sqlalchemy.engine import Engine

from .. import config
from zope.sqlalchemy import register

from hydra_base.exceptions import HydraError

import transaction
from sqlalchemy.orm import sessionmaker, declarative_base

import logging
log = logging.getLogger(__name__)

global DeclarativeBase
DeclarativeBase = declarative_base()

global DBSession
DBSession = None

global engine
engine = None

global hydra_db_url
hydra_db_url=None

global restart_counter
restart_counter = 0

#logger_sqlalchemy = logging.getLogger('sqlalchemy')
#logger_sqlalchemy.setLevel(logging.DEBUG)

# @event.listens_for(Engine, "connect")
# def set_sqlite_pragma(dbapi_connection, connection_record):
#     cursor = dbapi_connection.cursor()
#     cursor.execute("PRAGMA foreign_keys=ON")
#     cursor.close()

def create_mysql_db(db_url):
    """
        To simplify deployment, create the mysql DB if it's not there.
        Accepts a URL with or without a DB name stated, and returns a db url
        containing the db name for use in the main sqlalchemy engine.

        THe formats can take the following form:

        mysql+driver://username:password@hostname
        mysql+driver://username:password@hostname/dbname

        if no DB name is specified, it is retrieved from config
    """

    #add a special case for a memory-based sqlite session
    if db_url == 'sqlite://':
        return db_url

    #Remove trailing whitespace and forwardslashes
    db_url = db_url.strip().strip('/')


    #Check this is a mysql URL
    if db_url.find('mysql') >= 0:

        #Get the DB name from config and check if it's in the URL
        db_name = config.get('mysqld', 'db_name', 'hydradb')
        if db_url.find(db_name) >= 0:
            no_db_url = db_url.rsplit("/", 1)[0]
        else:
            #Check that there is a hostname specified, as we'll be using the '@' symbol soon..
            if db_url.find('@') == -1:
                raise HydraError("No Hostname specified in DB url")

            #Check if there's a DB name specified that's different to the one in config.
            host_and_db_name = db_url.split('@')[1]
            if host_and_db_name.find('/') >= 0:
                no_db_url, db_name = db_url.rsplit("/", 1)
            else:
                no_db_url = db_url
                db_url = no_db_url + "/" + db_name
        if db_url.find('charset') == -1:
            db_url = "{}?charset=utf8&use_unicode=1".format(db_url)

        if config.get('mysqld', 'auto_create', 'Y') == 'Y':
            tmp_engine = create_engine(no_db_url)
            log.debug("Creating database {0} as it does not exist.".format(db_name))
            with tmp_engine.connect() as conn:
                conn.execute(text("CREATE DATABASE IF NOT EXISTS {0}".format(db_name)))
    return db_url

def connect(db_url=None):
    if db_url is None:
        db_url = config.get('mysqld', 'url')

    log.info("Connecting to database")
    if db_url.find('@') >= 0:
        log.info("DB URL: %s", db_url.split('@')[1])
    else:
        log.info("DB URL: %s", db_url)

    db_url = create_mysql_db(db_url)

    global engine

    if db_url.startswith('sqlite'):
        engine = create_engine(db_url)
    else:

        # Let's use at least 10 for size and 20 for overflow (hydra.ini file)
        # To test the timeout: pool_size:1, max_overflow: 0, pool_timeout: 5 or any low value
        #These values MUST be smaller than the pool timeouts of the DB, otherwise the connection
        #will remain open on the client while it has been closed on the server, resulting in
        #an error
        db_pool_size = int(config.get('mysqld', 'pool_size',10)) # 10
        db_pool_recycle = int(config.get('mysqld', 'pool_recycle', 300)) # 300
        db_max_overflow = int(config.get('mysqld', 'max_overflow', 20)) # 10 -> 30
        db_pool_timeout = int(config.get('mysqld', 'pool_timeout', 10))
        db_pool_pre_ping = True if config.get('mysqld', 'pool_pre_ping', 'Y').upper() == 'Y' else False

        log.warning(f"db_pool_size: {db_pool_size} - pool_recycle: {db_pool_recycle} - max_overflow: {db_max_overflow} - pool_timeout: {db_pool_timeout} - pool_pre_ping: {db_pool_pre_ping}")

        engine = create_engine(db_url,
                               pool_recycle=db_pool_recycle,
                               pool_size=db_pool_size,
                               pool_timeout=db_pool_timeout,
                               max_overflow=db_max_overflow,
                               pool_pre_ping=db_pool_pre_ping)

    global hydra_db_url
    hydra_db_url=db_url


    global DBSession

    maker = sessionmaker(bind=engine, autoflush=False, autocommit=False)
    DBSession = scoped_session(maker)
    register(DBSession)

    global DeclarativeBase
    try:
        DeclarativeBase.metadata.create_all(engine, checkfirst=True)
    except sqlalchemy.exc.OperationalError as err:
        log.warning("Error creating database: %s", err)

    return db_url

def get_session():
    global DBSession
    return DBSession

def commit_transaction():
    try:
        transaction.commit()
    except Exception as e:
        log.critical(e)
        transaction.abort()

def open_session():
    log.debug("OPENING SESSION")

    global DBSession

    from .model import User
    session = DBSession()
    session.query(User).all()

    session2 = DBSession()
    session2.query(User).all()

    DBSession()

def close_session():
    log.debug("CLOSING SESSION")
    DBSession.remove()


def rollback_transaction():
    #import pudb; pudb.set_trace()
    transaction.abort()

def restart_session(caller='-- not specified --'):
    """
        WILL RESTART THE SESSION
    """
    global DBSession
    global restart_counter
    restart_counter = restart_counter + 1
    log.warning(f"[# Restarts: {restart_counter}] [{caller}] Restarting the DB Session!")
    close_session()
    global hydra_db_url
    connect(hydra_db_url)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from .base import *
from .permissions import User

from hydra_base.exceptions import PermissionError

__all__ = ['Rule', 'RuleTypeDefinition', 'RuleTypeLink']

class Rule(AuditMixin, Base, Inspect):
    """
        A rule is an arbitrary piece of text applied to either
        a Network, Project or Template.
    """

    __tablename__ = 'tRule'
    __table_args__ = (
        UniqueConstraint('network_id', 'name', name="unique network rule name"),
        UniqueConstraint('project_id', 'name', name="unique project rule name"),
        UniqueConstraint('template_id', 'name', name="unique template rule name")
    )

    __ownerfk__ = 'rule_id'

    id = Column(Integer(), primary_key=True, nullable=False)

    name = Column(String(200), nullable=False)
    description = Column(String(1000), nullable=True)

    format = Column(String(80), nullable=False, server_default='text')

    ref_key = Column(String(60), nullable=False, index=True)

    value = Column(Text().with_variant(mysql.LONGTEXT, 'mysql'), nullable=True)

    status = Column(String(1), nullable=False, server_default=text(u"'A'"))

    network_id = Column(Integer(), ForeignKey('tNetwork.id'), index=True, nullable=True)
    project_id = Column(Integer(), ForeignKey('tProject.id'), index=True, nullable=True)
    template_id = Column(Integer(), ForeignKey('tTemplate.id'), index=True, nullable=True)

    network = relationship('Network', backref=backref("rules",
                           order_by=network_id,
                           cascade="all, delete-orphan"),
                           lazy='joined')

    project = relationship('Project', backref=backref("rules",
                           order_by=project_id,
                           cascade="all, delete-orphan"),
                           lazy='joined')

    template = relationship('Template', backref=backref("templates",
                            order_by=project_id,
                            cascade="all, delete-orphan"),
                            lazy='joined')

    _parents = ['tProject', 'tNetwork', 'tTemplate']
    _children = []


    def set_types(self, types):
        """
            Accepts a list of type JSONObjects or spyne objects and sets
            the type of the rule to be exactly this. This means deleting rules
            which are not in the list
        """

        #We take this to mean don't touch types.
        if types is None:
            return

        existingtypes = set([t.code for t in self.types])

        #Map a type code to a type object
        existing_type_map = dict((t.code, t) for t in self.types)

        newtypes = set([t.code for t in types])

        types_to_add = newtypes - existingtypes
        types_to_delete = existingtypes - newtypes

        for ruletypecode in types_to_add:

            self.check_type_definition_exists(ruletypecode)

            ruletypelink = RuleTypeLink()
            ruletypelink.code = ruletypecode
            self.types.append(ruletypelink)

        for type_to_delete in types_to_delete:
            get_session().delete(existing_type_map[type_to_delete])

    def check_type_definition_exists(self, code):
        """
        A convenience function to check if a rule type definition exists before trying to add a link to it
        """

        try:
            get_session().query(RuleTypeDefinition).filter(RuleTypeDefinition.code == code).one()
        except NoResultFound:
            raise ResourceNotFoundError("Rule type definition with code {} does not exist".format(code))

    def get_network(self):
        """
        Rules are associated with a network directly or nodes/links/groups in a network,
        so rules are always associated to one network.
        This function returns that network
        """
        rule_network = None
        if self.ref_key.upper() == 'NETWORK':
            rule_network = self.network

        return rule_network


    @property
    def owners(self):
        rule_owners = []
        if self.network:
            rule_owners += self.network.get_owners()

        if self.project:
            rule_owners += self.project.get_owners()

        user_ids = set(o.user_id for o in rule_owners)
        return [{"user_id": user_id} for user_id in user_ids]


    def check_read_permission(self, user_id, do_raise=True):
        user = get_session().query(User).filter(User.id==user_id).one()
        if user.is_admin():
            return True

        if user_id in set(o["user_id"] for o in self.owners):
            return True
        else:
            if do_raise:
                raise PermissionError(f"user {user_id} does not have access to rule {self.id}")
            return False


    def check_write_permission(self, user_id, do_raise=True):
        if self.network:
            return self.network.check_write_permission(user_id, do_raise=do_raise)

        if self.project:
            return self.project.check_write_permission(user_id, do_raise=do_raise)

        if self.template:
            return self.template.check_write_permission(user_id, do_raise=do_raise)

        return False


    def asdict(self):
        """
         Dataclass-style dict representation for conversion to JSONObject
        """
        return {
          "id": self.id,
          "name": self.name,
          "value": self.value,
          "description": self.description,
          "status": self.status,
          "owners": self.owners
        }


    def get_owners(self):
        """
            Get all the owners of a rule, both those which are applied directly
            to this rule, but also who have been granted access via a project / network
        """
        return self.owners



class RuleTypeDefinition(AuditMixin, Base, Inspect):
    """
        Describes the types of rules available in the system

        A rule type is a simple way of categorising rules. A rule may have no
        type or it may have 1. A rule type consists of a unique code and a name.

        In addition to separating rules, this enables rules to be searched more easily.
    """

    __tablename__='tRuleTypeDefinition'

    __table_args__ = (
        UniqueConstraint('code', name="Unique Rule Code"),
    )

    code = Column(String(200), nullable=False, primary_key=True)
    name = Column(String(200), nullable=False)


class RuleTypeLink(AuditMixin, Base, Inspect):
    """
        Links rules to type definitions.

        A rule type is a simple way of categorising rules. A rule may have no
        type or it may have 1. A rule type consists of a unique code and a name.

        In addition to separating rules, this enables rules to be searched more easily.
    """

    __tablename__='tRuleTypeLink'

    __table_args__ = (
        UniqueConstraint('code', 'rule_id', name="Unique Rule / Type"),
    )

    code    = Column(String(200), ForeignKey('tRuleTypeDefinition.code'), primary_key=True, nullable=False)
    rule_id = Column(Integer(), ForeignKey('tRule.id'), primary_key=True, nullable=False)

    #Set the backrefs so that when a type definition or a rule are deleted, so are the links.
    typedefinition = relationship('RuleTypeDefinition', uselist=False, lazy='joined',
                                  backref=backref('ruletypes', cascade="all, delete-orphan"))
    rule = relationship('Rule', backref=backref('types', order_by=code, uselist=True, cascade="all, delete-orphan"))





"""unit_dimension_fk

Revision ID: 6402bb82a531
Revises: 5e381835e382
Create Date: 2019-03-18 15:07:16.318347

"""
from alembic import op
import sqlalchemy as sa

import logging
log = logging.getLogger(__name__)

# revision identifiers, used by Alembic.
revision = '6402bb82a531'
down_revision = '5e381835e382'
branch_labels = None
depends_on = None


def upgrade():

    try:
        op.add_column('tAttr', sa.Column('dimension_id', sa.Integer(), sa.ForeignKey('tDimension.id'), nullable=True))
        op.drop_column('tAttr', 'dimension')


        op.add_column('tTypeAttr', sa.Column('unit_id', sa.Integer(), sa.ForeignKey('tUnit.id'), nullable=True))
        op.drop_column('tTypeAttr', 'unit')

        op.add_column('tDataset', sa.Column('unit_id', sa.Integer(), sa.ForeignKey('tUnit.id'), nullable=True))
        op.drop_column('tDataset', 'unit')

    except Exception as e:
        log.exception(e)


    try:
        op.drop_constraint('unique name dimension', 'tAttr', type_='unique')
        op.create_unique_constraint('unique name dimension_id', 'tAttr', ['name', 'dimension_id'])
    except Exception as e:
        log.exception(e)

def downgrade():


    try:
        op.add_column('tAttr', sa.Column('dimension', sa.Integer, nullable=True))
        op.drop_column('tAttr', 'dimension_id')

        op.add_column('tTypeAttr', sa.Column('unit', sa.Integer(), nullable=True))
        op.drop_column('tTypeAttr', 'unit_id')

        op.add_column('tDataset', sa.Column('unit', sa.Integer(), nullable=True))
        op.drop_column('tDataset', 'unit_id')

    except Exception as e:
        log.exception(e)




#!/usr/bin/env python
# -*- coding: utf-8 -*-

# (c) Copyright 2013 to 2017 University of Manchester
#
# HydraPlatform is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HydraPlatform is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with HydraPlatform.  If not, see <http://www.gnu.org/licenses/>
#
from ..base import *

__all__ = ['ResourceGroupItem']

class ResourceGroupItem(Base, Inspect):
    """
    """

    __tablename__='tResourceGroupItem'

    __table_args__ = (
        UniqueConstraint('group_id', 'node_id', 'scenario_id', name='node_group_1'),
        UniqueConstraint('group_id', 'link_id', 'scenario_id',  name = 'link_group_1'),
        UniqueConstraint('group_id', 'subgroup_id', 'scenario_id', name = 'subgroup_group_1'),
    )

    id = Column(Integer(), primary_key=True, nullable=False)
    ref_key = Column(String(60),  nullable=False)

    node_id     = Column(Integer(),  ForeignKey('tNode.id'))
    link_id     = Column(Integer(),  ForeignKey('tLink.id'))
    subgroup_id = Column(Integer(),  ForeignKey('tResourceGroup.id'))

    group_id = Column(Integer(), ForeignKey('tResourceGroup.id'))
    scenario_id = Column(Integer(), ForeignKey('tScenario.id'),  nullable=False, index=True)

    cr_date = Column(TIMESTAMP(),  nullable=False, server_default=text(u'CURRENT_TIMESTAMP'))

    group = relationship('ResourceGroup', foreign_keys=[group_id], backref=backref("items", order_by=group_id))
    scenario = relationship('Scenario', backref=backref("resourcegroupitems", order_by=id, cascade="all, delete-orphan"))

    #These need to have backrefs to allow the deletion of networks & projects
    #--There needs to be a connection between the items & the resources to allow it
    node = relationship('Node', backref=backref("resourcegroupitems", order_by=id, cascade="all, delete-orphan"))
    link = relationship('Link', backref=backref("resourcegroupitems", order_by=id, cascade="all, delete-orphan"))
    subgroup = relationship('ResourceGroup', foreign_keys=[subgroup_id])

    _parents  = ['tResourceGroup', 'tScenario']
    _children = []

    def get_resource(self):
        ref_key = self.ref_key
        if ref_key == 'NODE':
            return self.node
        elif ref_key == 'LINK':
            return self.link
        elif ref_key == 'GROUP':
            return self.subgroup

    def get_resource_id(self):
        ref_key = self.ref_key
        if ref_key == 'NODE':
            return self.node_id
        elif ref_key == 'LINK':
            return self.link_id
        elif ref_key == 'GROUP':
            return self.subgroup_id
